"""Test deltacycle._kernel.finish"""

from deltacycle import Kernel, create_task, finish, get_kernel, run, sleep, step

from .conftest import trace


async def ctl():
    trace("enter")
    await sleep(100)

    # Force all PING threads to stop immediately
    trace("finish")
    finish()


async def ping(period: int):
    trace("enter")
    while True:
        await sleep(period)
        trace("PING")


EXP1 = {
    # CTL
    (0, "CTL", "enter"),
    (100, "CTL", "finish"),
    # FOO
    (0, "FOO", "enter"),
    (3, "FOO", "PING"),
    (6, "FOO", "PING"),
    (9, "FOO", "PING"),
    (12, "FOO", "PING"),
    (15, "FOO", "PING"),
    (18, "FOO", "PING"),
    (21, "FOO", "PING"),
    (24, "FOO", "PING"),
    (27, "FOO", "PING"),
    (30, "FOO", "PING"),
    (33, "FOO", "PING"),
    (36, "FOO", "PING"),
    (39, "FOO", "PING"),
    (42, "FOO", "PING"),
    (45, "FOO", "PING"),
    (48, "FOO", "PING"),
    (51, "FOO", "PING"),
    (54, "FOO", "PING"),
    (57, "FOO", "PING"),
    (60, "FOO", "PING"),
    (63, "FOO", "PING"),
    (66, "FOO", "PING"),
    (69, "FOO", "PING"),
    (72, "FOO", "PING"),
    (75, "FOO", "PING"),
    (78, "FOO", "PING"),
    (81, "FOO", "PING"),
    (84, "FOO", "PING"),
    (87, "FOO", "PING"),
    (90, "FOO", "PING"),
    (93, "FOO", "PING"),
    (96, "FOO", "PING"),
    (99, "FOO", "PING"),
    # BAR
    (0, "BAR", "enter"),
    (5, "BAR", "PING"),
    (10, "BAR", "PING"),
    (15, "BAR", "PING"),
    (20, "BAR", "PING"),
    (25, "BAR", "PING"),
    (30, "BAR", "PING"),
    (35, "BAR", "PING"),
    (40, "BAR", "PING"),
    (45, "BAR", "PING"),
    (50, "BAR", "PING"),
    (55, "BAR", "PING"),
    (60, "BAR", "PING"),
    (65, "BAR", "PING"),
    (70, "BAR", "PING"),
    (75, "BAR", "PING"),
    (80, "BAR", "PING"),
    (85, "BAR", "PING"),
    (90, "BAR", "PING"),
    (95, "BAR", "PING"),
    # FIZ
    (0, "FIZ", "enter"),
    (7, "FIZ", "PING"),
    (14, "FIZ", "PING"),
    (21, "FIZ", "PING"),
    (28, "FIZ", "PING"),
    (35, "FIZ", "PING"),
    (42, "FIZ", "PING"),
    (49, "FIZ", "PING"),
    (56, "FIZ", "PING"),
    (63, "FIZ", "PING"),
    (70, "FIZ", "PING"),
    (77, "FIZ", "PING"),
    (84, "FIZ", "PING"),
    (91, "FIZ", "PING"),
    (98, "FIZ", "PING"),
    # BUZ
    (0, "BUZ", "enter"),
    (11, "BUZ", "PING"),
    (22, "BUZ", "PING"),
    (33, "BUZ", "PING"),
    (44, "BUZ", "PING"),
    (55, "BUZ", "PING"),
    (66, "BUZ", "PING"),
    (77, "BUZ", "PING"),
    (88, "BUZ", "PING"),
    (99, "BUZ", "PING"),
}


def test_finish1(captrace: set[tuple[int, str, str]]):
    async def main():
        create_task(ctl(), name="CTL")
        create_task(ping(3), name="FOO")
        create_task(ping(5), name="BAR")
        create_task(ping(7), name="FIZ")
        create_task(ping(11), name="BUZ")

    # Subsequent calls to run() have no effect
    run(main())

    kernel = get_kernel()
    assert kernel is not None
    assert kernel.state() is Kernel.State.FINISHED

    assert captrace == EXP1


def test_finish2(captrace: set[tuple[int, str, str]]):
    async def main():
        create_task(ctl(), name="CTL")
        create_task(ping(3), name="FOO")
        create_task(ping(5), name="BAR")
        create_task(ping(7), name="FIZ")
        create_task(ping(11), name="BUZ")

    # Subsequent calls to run() have no effect
    list(step(main()))

    kernel = get_kernel()
    assert kernel is not None
    assert kernel.state() is Kernel.State.FINISHED
    assert kernel.done()

    assert captrace == EXP1



"""Test deltacycle.Semaphore"""

import pytest

from deltacycle import (
    AllOf,
    Lock,
    Semaphore,
    all_of,
    create_task,
    now,
    run,
    sleep,
)

from .conftest import trace


def test_len():
    async def main():
        sem = Semaphore(capacity=5)
        assert sem.capacity == 5
        assert len(sem) == 0
        sem.put()
        assert len(sem) == 1
        sem.put()
        assert len(sem) == 2
        assert sem
        await sem.get()
        assert len(sem) == 1
        await sem.get()
        assert len(sem) == 0
        assert not sem

    run(main())


async def use_get_put(sem: Semaphore, t1: int, t2: int):
    trace("enter")

    await sleep(t1)

    trace("attempt get")
    await sem.get()
    trace("acquired")

    try:
        await sleep(t2)
    finally:
        trace("put")
        sem.put()

    await sleep(10)
    trace("exit")


async def use_with(sem: Semaphore, t1: int, t2: int):
    trace("enter")

    await sleep(t1)

    trace("attempt get")
    async with sem.req():
        trace("acquired")
        await sleep(t2)
    trace("put")

    await sleep(10)
    trace("exit")


EXP = {
    # 0
    (0, "0", "enter"),
    (10, "0", "attempt get"),
    (10, "0", "acquired"),
    (20, "0", "put"),
    (30, "0", "exit"),
    # 1
    (0, "1", "enter"),
    (11, "1", "attempt get"),
    (11, "1", "acquired"),
    (21, "1", "put"),
    # 2
    (0, "2", "enter"),
    (12, "2", "attempt get"),
    (12, "2", "acquired"),
    (22, "2", "put"),
    (32, "2", "exit"),
    # 3
    (0, "3", "enter"),
    (13, "3", "attempt get"),
    (13, "3", "acquired"),
    (23, "3", "put"),
    (33, "3", "exit"),
    # 4
    (0, "4", "enter"),
    (14, "4", "attempt get"),
    (20, "4", "acquired"),
    (30, "4", "put"),
    (40, "4", "exit"),
    # 5
    (0, "5", "enter"),
    (15, "5", "attempt get"),
    (21, "5", "acquired"),
    (31, "5", "put"),
    (41, "5", "exit"),
    # 6
    (0, "6", "enter"),
    (16, "6", "attempt get"),
    (22, "6", "acquired"),
    (32, "6", "put"),
    (42, "6", "exit"),
    # 7
    (0, "7", "enter"),
    (17, "7", "attempt get"),
    (23, "7", "acquired"),
    (31, "1", "exit"),
    (33, "7", "put"),
    (43, "7", "exit"),
}


def test_get_put(captrace: set[tuple[int, str, str]]):
    async def main():
        sem = Semaphore(4)
        for i in range(8):
            create_task(use_get_put(sem, i + 10, 10), name=f"{i}")

    run(main())

    assert captrace == EXP


def test_async_with(captrace: set[tuple[int, str, str]]):
    async def main():
        sem = Semaphore(4)
        for i in range(8):
            create_task(use_with(sem, i + 10, 10), name=f"{i}")

    run(main())

    assert captrace == EXP


def test_unbounded():
    async def use_unbounded():
        sem = Semaphore(2)

        await sem.get()
        await sem.get()
        sem.put()
        sem.put()

        # No exception!
        sem.put()

    run(use_unbounded())


def test_bounded():
    async def use_bounded():
        sem = Semaphore(value=2, capacity=2)

        await sem.get()
        await sem.get()

        sem.put()
        sem.put()

        # Exception!
        with pytest.raises(OverflowError):
            sem.put()

    async def main():
        create_task(use_bounded())

    run(main())


def test_priority():
    async def request(lock: Lock, p: int):
        await sleep(5)
        await lock.get(priority=p)
        await sleep(5)
        lock.put()

    async def main():
        lock = Lock()
        t3 = create_task(request(lock, 3), name="T3")
        t2 = create_task(request(lock, 2), name="T2")
        t1 = create_task(request(lock, 1), name="T1")
        t0 = create_task(request(lock, 0), name="T0", priority=-1)

        ts = await all_of(t3, t2, t1, t0)
        assert ts == (t0, t1, t2, t3)
        assert now() == 25

    run(main())


def test_init_bad_values():
    with pytest.raises(ValueError):
        _ = Semaphore(value=5, capacity=4)

    with pytest.raises(ValueError):
        _ = Semaphore(-1)


def test_schedule_all1():
    async def cf(s: Semaphore, t1: int, t2: int, t3: int):
        await sleep(t1)
        await s.get()
        await sleep(t2)
        s.put()
        await sleep(t3)

    async def main():
        lock = Lock()
        t1 = create_task(cf(lock, 0, 10, 10))

        await sleep(1)
        ys = await AllOf(t1, lock.req())

        assert ys == (lock, t1)
        assert now() == 20
        assert not lock
        lock.put()

    run(main())


def test_schedule_all2():
    async def cf(s: Semaphore, t1: int, t2: int, t3: int):
        await sleep(t1)
        await s.get()
        await sleep(t2)
        s.put()
        await sleep(t3)

    async def main():
        lock = Lock()
        t1 = create_task(cf(lock, 0, 10, 10))

        await sleep(1)
        ys = await all_of(t1, lock.req())

        assert ys == (lock, t1)
        assert now() == 20
        assert not lock
        lock.put()

    run(main())



"""Test bugs"""

from deltacycle import TaskGroup, run, sleep

from .common import Bool
from .conftest import trace

EXP2 = {
    (5, "do_stuff", "first"),
    (15, "do_stuff", "second"),
    (25, "do_stuff", "third"),
    (35, "do_stuff", "fourth"),
}


def test_2(captrace: set[tuple[int, str, str]]):
    clock = Bool(name="clock")

    async def do_stuff():
        await clock.posedge()
        trace("first")
        await clock.posedge()
        trace("second")
        await clock.posedge()
        trace("third")
        await clock.posedge()
        trace("fourth")

    async def drv_clock():
        clock.next = False
        while True:
            await sleep(5)
            clock.next = not clock.value

    async def main():
        async with TaskGroup() as tg:
            tg.create_task(drv_clock(), name="drv_clock")
            tg.create_task(do_stuff(), name="do_stuff")

    run(main(), until=100)

    assert captrace == EXP2



"""Simulate a 4-bit adder."""

from deltacycle import any_of, create_task, run, sleep

from .common import Bool
from .conftest import trace

# a, b, ci, s, co
VALS = [
    (False, False, False, False, False),
    (False, False, True, True, False),
    (False, True, False, True, False),
    (False, True, True, False, True),
    (True, False, False, True, False),
    (True, False, True, False, True),
    (True, True, False, False, True),
    (True, True, True, True, True),
]

EXP = {
    (10 * i + 5, "mon_outputs", f"s={VALS[i][3]:b} co={VALS[i][4]:b}") for i, _ in enumerate(VALS)
}


def test_add(captrace: set[tuple[int, str, str]]):
    """Test 4-bit adder simulation."""
    period = 10

    # Inputs
    clk = Bool(name="clk")

    a = Bool(name="a")
    b = Bool(name="b")
    ci = Bool(name="ci")

    # Outputs
    s = Bool(name="s")
    co = Bool(name="co")

    async def drv_clk():
        clk.next = False
        while True:
            await sleep(period // 2)
            clk.next = not clk.prev

    async def drv_inputs():
        for a_val, b_val, ci_val, _, _ in VALS:
            a.next = a_val
            b.next = b_val
            ci.next = ci_val
            await clk.posedge()

    async def drv_outputs():
        while True:
            await any_of(a, b, ci)
            g = a.value & b.value
            p = a.value | b.value
            s.next = a.value ^ b.value ^ ci.value
            co.next = g | p & ci.value

    async def mon_outputs():
        while True:
            await clk.posedge()
            trace(f"s={s.prev:b} co={co.prev:b}")

    async def main():
        create_task(drv_clk(), name="drv_clk", priority=0)
        create_task(drv_inputs(), name="drv_inputs", priority=0)
        create_task(drv_outputs(), name="drv_outputs", priority=-1)
        create_task(mon_outputs(), name="mon_outputs", priority=1)

    until = period * len(VALS)
    run(main(), until=until)

    assert captrace == EXP



"""Hello, world!"""

import pytest

from deltacycle import Kernel, get_kernel, run, sleep

from .conftest import trace

EXP = {
    (-1, "", "Before Time"),
    (2, "main", "Hello"),
    (4, "main", "World"),
}


def test_hello(captrace: set[tuple[int, str, str]]):
    """Test basic async/await hello world functionality."""
    trace("Before Time")

    async def hello():
        await sleep(2)
        trace("Hello")
        await sleep(2)
        trace("World")
        return 42

    ret = run(hello())
    assert ret == 42

    kernel = get_kernel()
    assert kernel is not None
    assert kernel.state() is Kernel.State.COMPLETED

    with pytest.raises(RuntimeError):
        run(kernel=kernel)

    assert captrace == EXP



"""Test seqlogic.sim.Lock class."""

from deltacycle import Lock, create_task, run, sleep

from .conftest import trace


async def use_acquire_release(lock: Lock, t1: int, t2: int):
    trace("enter")

    await sleep(t1)

    trace("attempt get")
    await lock.get()
    trace("acquired")

    try:
        await sleep(t2)
    finally:
        trace("put")
        lock.put()

    await sleep(10)
    trace("exit")


async def use_with(lock: Lock, t1: int, t2: int):
    trace("enter")

    await sleep(t1)

    trace("attempt get")
    async with lock.req():
        trace("acquired")
        await sleep(t2)
    trace("put")

    await sleep(10)
    trace("exit")


EXP = {
    (0, "0", "enter"),
    (0, "1", "enter"),
    (0, "2", "enter"),
    (0, "3", "enter"),
    (10, "0", "attempt get"),
    (10, "0", "acquired"),
    (11, "1", "attempt get"),
    (12, "2", "attempt get"),
    (13, "3", "attempt get"),
    (20, "0", "put"),
    (20, "1", "acquired"),
    (30, "0", "exit"),
    (30, "1", "put"),
    (30, "2", "acquired"),
    (40, "1", "exit"),
    (40, "2", "put"),
    (40, "3", "acquired"),
    (50, "2", "exit"),
    (50, "3", "put"),
    (60, "3", "exit"),
}


def test_acquire_release(captrace: set[tuple[int, str, str]]):
    async def main():
        lock = Lock()
        for i in range(4):
            create_task(use_acquire_release(lock, i + 10, 10), name=f"{i}")

    run(main())

    assert captrace == EXP


def test_async_with(captrace: set[tuple[int, str, str]]):
    async def main():
        lock = Lock()
        for i in range(4):
            create_task(use_with(lock, i + 10, 10), name=f"{i}")

    run(main())

    assert captrace == EXP



"""Simulate a register file."""

from deltacycle import any_of, create_task, run, sleep

from .common import Bool, Int, IntMem
from .conftest import trace

# wr_en, wr_addr, wr_data, rd_addr, rd_data
VALS = [
    (True, 0, 42, 0, 0),
    (True, 1, 43, 0, 42),
    (True, 2, 44, 1, 43),
    (True, 3, 45, 2, 44),
    (True, 4, 46, 3, 45),
    (True, 5, 47, 4, 46),
    (True, 6, 48, 5, 47),
    (True, 7, 49, 6, 48),
    (True, 8, 50, 7, 49),
    (True, 9, 51, 8, 50),
    # Note: Same WrData
    (True, 0, 42, 9, 51),
    (False, 0, 0, 0, 42),
]

EXP = {
    (
        10 * i + 5,
        "mon_outputs",
        f"wr_en={VALS[i][0]:b} wr_addr={VALS[i][1]:x} wr_data={VALS[i][2]:02x} "
        f"rd_addr={VALS[i][3]:x} rd_data={VALS[i][4]:02x}",
    )
    for i, _ in enumerate(VALS)
}


def test_regfile(captrace: set[tuple[int, str, str]]):
    clk = Bool(name="clk")
    period = 10

    wr_en = Bool(name="wr_en")
    wr_addr = Int(name="wr_addr")
    wr_data = Int(name="wr_data")

    rd_addr = Int(name="rd_addr")
    rd_data = Int(name="rd_data")

    # State
    regs = IntMem(name="regs")

    async def drv_clk():
        clk.next = False
        while True:
            await sleep(period // 2)
            clk.next = not clk.prev

    async def drv_inputs():
        for we, wa, wd, ra, _ in VALS:
            wr_en.next = we
            wr_addr.next = wa
            wr_data.next = wd
            rd_addr.next = ra
            await clk.posedge()

    async def mon_outputs():
        while True:
            await clk.posedge()
            trace(
                f"wr_en={wr_en.prev:b} wr_addr={wr_addr.prev:x} wr_data={wr_data.prev:02x} "
                f"rd_addr={rd_addr.prev:x} rd_data={rd_data.prev:02x}"
            )

    async def wr_port():
        def clk_pred():
            return clk.is_posedge() and wr_en.prev

        while True:
            await clk.pred(clk_pred)
            regs[wr_addr.prev].next = wr_data.prev

    async def rd_port():
        while True:
            await any_of(regs, rd_addr)
            rd_data.next = regs.value[rd_addr.value]

    async def main():
        create_task(drv_clk(), name="drv_clk", priority=2)
        create_task(drv_inputs(), name="drv_inputs", priority=2)
        create_task(mon_outputs(), name="mon_outputs", priority=3)
        create_task(wr_port(), name="wr_port", priority=2)
        create_task(rd_port(), name="rd_port", priority=1)

    run(main(), until=120)

    for i in range(10):
        assert regs[i].prev == 42 + i

    assert captrace == EXP
